// ==============================================================================
// Test fuer einen evolutionaeren Algorithmus. 
// Aufgabe: Wegfindung. Die Individuen koennen sich in jedem Frame beliebig drehen. 
// Version 0.6 - Demo, Jan Kube, Dez. 2017


// VARIABLEN ====================================================================
// Simulation
PImage hintergrund;                                          // enthaelt spaeter den Parkours
int frameZaehler = 0;                                        // Zaehlt, wie viele Frames schon simuliert wurden
int frameZaehlerGrenze = 1200;                               // wie viele Frames sollen die Individuen leben?
boolean alleTot = true;                                      // wenigstens einer muss sich hier als lebend melden
// Ziel
ArrayList<Punkt> zielPunkte = new ArrayList<Punkt>();        // Da wollen alle hin (Element 1, 2 und 3) - Nahrung
int zielGroesse = 80;                                        // so gross ist jedes Ziel (Nahrung)
int punkteProNahrungsquelle = 1000;                          // So viele Punkte gibts bei Erreichen eines neuen Ziels (Nahrungsquelle)
int punkteAbzugBeiCrash = 1000;
int zielPunkteAdderCooldown = 0;                             // nur langsam neue Zielpunkte einfuegen koennen, nicht zu viel auf einmal
// Individuum
Punkt spawnPunkt = new Punkt(120, 415, 90);                   // Ausgangspunkt fuer alle Individuen (inkl. Winkel)
int individuumsGroesse = 10;                                 // so gross ist das Individuum
float laengeSchrittMaximal = 10;                             // so weit kommt ein Individuum pro Frame
int maxDrehungProFrame = 45;                                 // maximale Drehung pro Frame
int maxSichtweite = 50;                                      // wie weit koennen die Sensoren maximal sehen?
int sensorAnzahl = 5;                                        // so viele Sensoren gibt es
// Evolution
int generationZaehler = 1;                                   // welche Generation haben wir gerade?
int spawnZahl = 100;                                         // Individuen pro Generation
float prozentVererbung = 80;                                 // wie viel Prozent der (besten) Population duerfen sich vermehren?
float prozentBesteSafe = 5;                                  // so viel Prozent der besten Individuen sind sicher in neuer Individuen-Liste
float prozentMutation = 5;                                   // wie viel Mutation soll auftreten
float prozentTotaleMutation = 10;                            // in so viel Prozent der Faelle soll ein neues, zufaelliges Individuum entstehen
ArrayList<Individuum> individuen = new ArrayList<Individuum>();  // Liste aller Individuen





// PUNKT-KLASSE =================================================================
class Punkt {
  float x;
  float y;
  float drehung;
  
  // Constructor ohne Drehungsangabe
  public Punkt(float tx, float ty) {
    this.x = tx;
    this.y = ty;
    this.drehung = 0;
  }
  // Constructor mit Drehungsangabe
  public Punkt(float tx, float ty, float td) {
    this.x = tx;
    this.y = ty;
    this.drehung = td; // % 360; // 370 Grad entsprechen 10 Grad, also reicht es, 10 zu speichern (Rest bei Teilung durch 360)
  }
}


// SENSOR-KLASSE ================================================================
class Sensor {
  // Aufbau
  float winkel; // in welchem Winkel ist der Sensor am Koerper befestigt?
  float sichtweite;
  // Reaktion
  float drehungBeiKontakt;
  
  // Constructor ohne Argumente (bei initialer Erstellung)
  public Sensor() {
    this.winkel = random(360);
    this.sichtweite = random(maxSichtweite);
    this.drehungBeiKontakt = (random(maxDrehungProFrame*2)-maxDrehungProFrame); // + und - Drehung erlauben
  }
  // Constructor bei allen mitgegebenen Argumenten (fuer Vererbung)
  public Sensor(int w, int s, int d) {
    this.winkel = w;
    this.sichtweite = s;
    this.drehungBeiKontakt = d;
    // Nicht zu viel Drehung pro Frame zulassen
    if (this.drehungBeiKontakt >      maxDrehungProFrame) { this.drehungBeiKontakt =      maxDrehungProFrame; }
    if (this.drehungBeiKontakt < (-1)*maxDrehungProFrame) { this.drehungBeiKontakt = (-1)*maxDrehungProFrame; }
  }
  
  // Zeichnen (als Linie von Punkt p in Richtung der Drehung von Punkt p in Laenge der Sichtweite dieses Sensors)
  void zeichnen(Punkt p) {
    Punkt tempSensorPunkt = new Punkt(p.x, p.y, p.drehung + this.winkel); // Die Winkel-Abweichung dieses speziellen Sensors zum Winkel des Individuums hinzuaddieren
    Punkt tempZiel = berechneZielpunktAusPositionUndWinkel(tempSensorPunkt, this.sichtweite);
    line(p.x, p.y, tempZiel.x, tempZiel.y);
  }
  
  // Sieht der Sensor etwas?
  boolean seheHindernis(Punkt p) {
    // Pruefen, ob sich der Zielpunkt des Sensors in schwarzem Bereich befindet -> Hindernis gesehen
    Punkt tempSensorPunkt = new Punkt(p.x, p.y, p.drehung + this.winkel); // Die Winkel-Abweichung dieses speziellen Sensors zum Winkel des Individuums hinzuaddieren
    Punkt sichtZiel = berechneZielpunktAusPositionUndWinkel(tempSensorPunkt, this.sichtweite);
    if (hintergrund.get((int)sichtZiel.x, (int)sichtZiel.y) == color(0,0,0)) {
      return true; // Schwarz
    } else {
      return false; // kein Schwarz
    }
  }
  
  float drehReaktion(Punkt p) { //<>//
    if (this.seheHindernis(p)) {
      return this.drehungBeiKontakt; //<>//
    } else {
      return 0; // TODO: Auch eine Dreh-Reaktion zeigen koennen, wenn nichts in Sichtweite ist. 
    }
  }
  
  Sensor mischen(Sensor s) {
    // Hier werden zwei Sensoren gemischt: Dieser und ein mitgelieferter. Wichtig fuer Vererbung. 
    Sensor tempSensor = new Sensor();
    tempSensor.winkel            = (this.winkel + s.winkel) / 2;   // Winkel mischen
    tempSensor.sichtweite        = (this.sichtweite + s.sichtweite) / 2;   // Sichtweiten mischen
    tempSensor.drehungBeiKontakt = ((this.drehungBeiKontakt + s.drehungBeiKontakt) / 2) % 360;   // Drehungs-Reaktionen mischen
    
    // Mutationen der Sensoren
    // Winkel
    float schwankungsBereichWinkel = 360 * (prozentMutation / 100.0);
    float zufallsZahlWinkel = random(schwankungsBereichWinkel) - (schwankungsBereichWinkel/2.0); // auch negative Schwankung erlauben
    tempSensor.winkel += zufallsZahlWinkel;
    // Sichtweite
    float schwankungsBereichSichtweite = maxSichtweite * (prozentMutation / 100.0);
    float zufallsZahlSichtweite = random(schwankungsBereichSichtweite) - (schwankungsBereichSichtweite/2.0); // auch negative Schwankung erlauben
    tempSensor.sichtweite += zufallsZahlSichtweite;
    // DrehungBeiKontakt
    float schwankungsBereichDrehung = 360 * (prozentMutation / 100.0);
    float zufallsZahlDrehung = random(schwankungsBereichDrehung) - (schwankungsBereichDrehung/2.0); // auch negative Schwankung erlauben
    tempSensor.drehungBeiKontakt += zufallsZahlDrehung;
    
    // Rueckgabe
    return tempSensor;
  }
}


// INDIVIDUUM-KLASSE ============================================================
class Individuum {
  // Position
  Punkt position; // x, y und drehung
  float groesse;
  // Fitness
  boolean raus;
  boolean crash;
  boolean[] getroffen;
  float punkte;
  // Sensoren
  ArrayList<Sensor> sensoren;
  // weitere Eigenschaften
  float laengeSchritt;
  
  // Constructor
  Individuum(Punkt tp) {
    this.position = tp;
    this.groesse = individuumsGroesse; // TODO: Variabel machen (auch die Vererbung nicht vergessen)
    this.getroffen = new boolean[100];
    this.raus = false;
    this.crash = false;
    this.punkte = 0.0;
    sensoren = new ArrayList<Sensor>();
    for (int i=0; i<sensorAnzahl; i++) {
      sensoren.add(new Sensor());
    }
    this.laengeSchritt = random(laengeSchrittMaximal-1)+1;
  }
  
  // Methoden
  void pruefeAufZiel() {
    // Alle Ziele durchgehen
    for (int i=0; i<zielPunkte.size(); i++) {
      if (this.position.x > (zielPunkte.get(i).x-zielGroesse/2)   &&   this.position.x < (zielPunkte.get(i).x+zielGroesse/2)) {    // X-Pruefung
        if (this.position.y > (zielPunkte.get(i).y-zielGroesse/2)   &&   this.position.y < (zielPunkte.get(i).y+zielGroesse/2)) {  // Y-Pruefung
          if (this.getroffen[i] == false) {  // nur Punkte vergeben, wenn bisher noch nicht geschehen
            this.getroffen[i] = true;
            this.punkte += punkteProNahrungsquelle;
          }
        }
      }
    }
  }
  
  void pruefeAufRand() {
    // nur pruefen, wenn noch nicht raus
    if (!this.raus) {  
      // Pruefung X oder Y
      if (this.position.x > width-this.groesse || 
          this.position.x < this.groesse || 
          this.position.y > height-this.groesse || 
          this.position.y < this.groesse) { 
        raus = true;
      }
    }
  }
  
  void pruefeAufKollision() {
    if (!this.crash) {  // nur pruefen, wenn noch kein Unfall passiert ist
      // Pruefen, ob sich das Individuum auf Schwarz befindet -> crash
      if (hintergrund.get((int)this.position.x, (int)this.position.y) == color(0,0,0)) {
        this.crash = true;
        /*
        this.punkte -= punkteAbzugBeiCrash;
        if (this.punkte < 0) {
          this.punkte = 0;
        }
        */
      }
    }
  }
  
  void drehen() { //<>//
    // nur bewegen, wenn noch nicht raus oder verunfallt
    if (!this.raus && !this.crash) {   //<>//
      // Drehen - Mischung aus allen Drehungsangaben aller Sensoren
      int tempDrehung = 0; //<>//
      // Dreh-Reaktionen der Sensoren auf die momentane Individuums-Position aufaddieren
      for (int i=0; i<this.sensoren.size(); i++) {
        tempDrehung += this.sensoren.get(i).drehReaktion(this.position); //<>//
      }
       //<>// //<>//
      // Drehen
      //println("Sensor 1: " + sensorLinks.drehReaktion(this.position) + ", Sensor 2: " + sensorMitte.drehReaktion(this.position) + ", Sensor 3: " + sensorRechts.drehReaktion(this.position) + ", Summe: " + tempDrehung);
      // Drehung(saenderung) merken; jeweils aber nur den Rest nach 360er Teilung behalten. 
      tempDrehung = tempDrehung % 360; //<>//
      this.position.drehung += tempDrehung; //<>//
      this.position.drehung = this.position.drehung % 360; //<>//
    }
  }
  
  void bewegen() {
    // nur bewegen, wenn noch nicht raus oder verunfallt
    if (!this.raus && !this.crash) {
      // Bewegen - Zielpunkt fuer Bewegung in Blickrichtung berechnen
      this.position = berechneZielpunktAusPositionUndWinkel(this.position, this.laengeSchritt);
      // Bewegung speichern
      
      // Punkt gewaehren
      this.punkte+=0.1;
      
      // mich als lebend melden
      alleTot = false;
    }
  }
  
  void zeichnen() {
    // 1. Einfaerbung aendern (fuer erfolgreiche Individuen)? (Mit jedem Treffer wird das Individuum immer gruener)
    int trefferzahl = 0;
    for (int i=0; i<zielPunkte.size(); i++) {
      if (this.getroffen[i]==true) {
        trefferzahl++;
      }
    }
    if (trefferzahl == 0) { fill(255,0,0); }     // rot
    if (trefferzahl == 1) { fill(255,255,0); }   // gelb
    if (trefferzahl == 2) { fill(0,255,0); }     // gruen
    if (trefferzahl == 3) { fill(0,255,255); }   // cyan
    if (trefferzahl == 4) { fill(255,255,255); } // weiss
    if (trefferzahl == 5) { fill(255,0,255); }   // pink
    if (trefferzahl == 6) { fill(0,0,0); }       // schwarz
    
    // 2. Koerper zeichnen
    ellipse(this.position.x, this.position.y, this.groesse, this.groesse);
    
    // 3. Sensor-Ausleger zeichnen
    for (int i=0; i<this.sensoren.size(); i++) {
      this.sensoren.get(i).zeichnen(this.position);
    }
    
  }
  
  Individuum erzeugeNachkommen(Individuum paarungsPartner) {
    // neues Individuum erzeugen (randomisiert)
    Individuum tempInd = new Individuum(spawnPunkt);
    
    // pruefen, ob der TOTALE_MUTATION-Fall vorliegt: dann nicht mischen, sondern so zufaellig lassen
    if (random(1) < prozentTotaleMutation/100.0) {
      // nichts tun
    } else {
      // Alle existenten Sensoren loeschen
        tempInd.sensoren.clear();
      // Vererbungsfall: die Sensor-Eigenschaften der zwei Eltern vermischen und mitteln, dann in Kind-Sensor speichern
      for (int i=0; i<this.sensoren.size(); i++) {
        tempInd.sensoren.add( this.sensoren.get(i).mischen(paarungsPartner.sensoren.get(i)) );
      }
      // schrittlaenge uebernehmen und mutieren
      tempInd.laengeSchritt = (this.laengeSchritt + paarungsPartner.laengeSchritt) / 2.0; // Vererben...      
      float schwankungsBereichSchrittlaenge = laengeSchrittMaximal * (prozentMutation / 100.0); // und mutieren
      float zufallsZahlSchrittlaenge = random(schwankungsBereichSchrittlaenge) - (schwankungsBereichSchrittlaenge/2.0); // auch negative Schwankung erlauben
      tempInd.laengeSchritt += zufallsZahlSchrittlaenge; // Mutation aufrechnen
      // Sanity Check fuer Schrittlaenge
      if (tempInd.laengeSchritt > laengeSchrittMaximal) {
        tempInd.laengeSchritt = laengeSchrittMaximal;
      } else if (tempInd.laengeSchritt < 1) {
        tempInd.laengeSchritt = 1;
      }
    }
    
    // Ergebnis (Kind-Individuum) zurueckggeben
    return tempInd;
  }
}



// METHODEN ============================================================
void setup() {
  size(1200, 800);
  frameRate(60);
  erzeugeAlleIndividuenInitial();
  hintergrund = loadImage("hoehle.png");
  hintergrund.resize(width, height);
  hintergrund.loadPixels();
  
  // Ziele (Nahrung hinzufuegen)
  zielPunkte.add(new Punkt(220,450));
  zielPunkte.add(new Punkt(660,610));
  zielPunkte.add(new Punkt(900,390));
  zielPunkte.add(new Punkt(1100,700));
}

Punkt berechneZielpunktAusPositionUndWinkel(Punkt p, float l) {
  // Aus einem Startpunkt mitsamt einem Winkel und einer Entfernung wird der Zielpunkt berechnet
  float tempZielX = sin(radians(p.drehung)) * l + p.x;
  float tempZielY = cos(radians(p.drehung)) * l + p.y; 
  // Zielpunkt sollt innerhalb des Zeichenbereiches liegen -> beschraenken
  if (tempZielX > width) { tempZielX = width; }
  if (tempZielX < 0)     { tempZielX = 0; }
  if (tempZielY > height) { tempZielY = height; }
  if (tempZielY < 0)     { tempZielY = 0; }
  // Rueckgabe (der Winkel bleibt)
  return (new Punkt(tempZielX, tempZielY, p.drehung));
}

void erzeugeAlleIndividuenInitial() {
  for(int i=0; i<spawnZahl; i++) {
    // Temp-Individuum erzeugen und in Liste einfuegen (zufaelliger Winkel)
    Individuum tempInd = new Individuum(spawnPunkt);
    individuen.add(tempInd);
  }
}

void zeichneAlleIndividuen() {
  // Aussehen festlegen
  fill(255, 255, 0);
  stroke(0);
  strokeWeight(1);
  
  // Alle Individuen durchgehen
  for(int i=0; i<individuen.size(); i++) {
    individuen.get(i).zeichnen();
  }
}

void bewegeAlleIndividuen() {
  for(int i=0; i<individuen.size(); i++) {
    // Auf Treffen pruefen
    individuen.get(i).pruefeAufZiel();
    individuen.get(i).pruefeAufRand();
    individuen.get(i).pruefeAufKollision();
    // Bewegung ausloesen
    individuen.get(i).drehen();
    individuen.get(i).bewegen();
    // Wenn niemand sich als lebend meldete, dann Ende (neue Generation bitte)
    if (alleTot) {
      frameZaehler = frameZaehlerGrenze - 1;
    }
  }
}

void zeichneZielPunkte() {
  fill(0,0,255);
  for (int i=0; i<zielPunkte.size(); i++) {
    ellipse(zielPunkte.get(i).x, zielPunkte.get(i).y, zielGroesse, zielGroesse);
  }
}

void vererben() {
  // neue Generation -> wieder auf Tod pruefen
  alleTot = true; 
  
  // Zaehler zurueck-/hochsetzen
  frameZaehler = 0;
  generationZaehler++;
  
  /* TODO
  // So viele der besten Individuen haben schon einen reservierten Platz, der ist schon weg: 
  ArrayList<Individuum> besteKopieListe = new ArrayList<Individuum>();
  int anzahlSafeIndividuen = (int)(prozentBesteSafe * spawnZahl / 100.0); 
  println(anzahlSafeIndividuen);
  // Die besten werden gerettet!
  ArrayList<Individuum> besteSafeListe = new ArrayList<Individuum>();
  for (int i=0; i<anzahlSafeIndividuen; i++) {
    int tempIndexMax = findeIndexMaximalerPunkte();
    besteSafeListe.add(individuen.get(tempIndexMax));
  }
  */

  // neue Liste erstellen (zum Vermehren und Vererben)
  ArrayList<Individuum> kopulationsListe = new ArrayList<Individuum>();
  // X Prozent der bisher besten Elemente finden
  for (int i=0; i<prozentVererbung*individuen.size()/100.0; i++) {
    int tempIndexMax = findeIndexMaximalerPunkte();
    // Individuum so oft in Kopulations-Liste uebernehmen, wie es Punkte hat (um Vererbungschancen zu erhoehen) // TODO: Bessere Technik als total oft einfuegen finden
    for (int k=0; k<(individuen.get(tempIndexMax)).punkte; k++) {
      kopulationsListe.add(individuen.get(tempIndexMax));
    }
    // Individuum aus bisheriger Liste entfernen
    individuen.remove(tempIndexMax);
  }
  
  // alte Liste der Individuen loeschen
  individuen = new ArrayList<Individuum>();
  individuen.clear();
  
  // aus Kopulationsliste werden Nachkommen erzeugt
  for (int k=0; k<spawnZahl; k++) {
    // zufaellige Elternteile aus Kopulationsliste waehlen und Nachkommen erzeugen lassen
    Individuum elternteil_1 = kopulationsListe.get( (int)random(kopulationsListe.size()) );
    Individuum elternteil_2 = kopulationsListe.get( (int)random(kopulationsListe.size()) );
    // Nachkommen erstellen lassen (aus Mischung beider Elternteile)
    Individuum nachkomme = elternteil_1.erzeugeNachkommen(elternteil_2);
    // Nachkommen mit vererbter, mutierter Sensorreaktion in regulaere Individuen-Liste einfuegen
    individuen.add(nachkomme);
  }
  
  /*
  // die anfangs "gesafeten" Individuen werden in die Populationsliste zurueckgeschrieben
  for (int i=0; i<besteSafeListe.size(); i++) {
    individuen.add(besteSafeListe.get(i));
  }
  */
}

int findeIndexMaximalerPunkte() {
  // das jeweilige Element mit maximaler Punktzahl finden
  int maxPunkteIndex = 0;
  float maxPunkte = 0;
  for (int j=0; j<individuen.size(); j++) {
    if (individuen.get(j).punkte > maxPunkte) {
      maxPunkteIndex = j;   // neues Maximum gefunden -> Index merken
      maxPunkte = individuen.get(j).punkte;   // neuen Hoechst-Punktestand merken
    }
  }
  // Ergebnis liefern
  return maxPunkteIndex;
}



// DRAW ============================================================
void draw() {
  /*
  // Debug
  print("Alle Individuen:");
  for(int i=0; i<individuen.size(); i++) {
    println("Nr. " + i + ": Drehung " + individuen.get(i).position.drehung);
  }
  */
  
  // Zilpunkte-Hinzufuegen: Cooldownzeit herunterzaehlen
  if (zielPunkteAdderCooldown > 0) {
    zielPunkteAdderCooldown--;
  }
  // Mit Mausklick eine Nahrungsquelle einfuegen
  if (mousePressed && zielPunkteAdderCooldown==0) {
    zielPunkte.add(new Punkt(mouseX, mouseY));
    zielPunkteAdderCooldown=(int)frameRate;
  }
  
  // Zeichnen
  background(hintergrund);
  zeichneZielPunkte();
  zeichneAlleIndividuen();
  fill(255,0,0);
  text("Evolutions-Wegfindungs-Test. Generation " + generationZaehler + ", Frame " + frameZaehler + "/" + frameZaehlerGrenze, 10,20);
  
  // Simulieren
  bewegeAlleIndividuen();
  
  // Genug fuer diese Generation?
  frameZaehler++;
  if (frameZaehler >= frameZaehlerGrenze) {
    vererben();
  }
}
 